/*  This file is part of the KI18N Framework
    SPDX-FileCopyrightText: 2014 Kevin Krammer <krammer@kde.org>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#ifndef TESTHELPERS_H
#define TESTHELPERS_H

#include <QHash>

bool deployTestConfig();
bool removeTestConfig();

#endif // TESTHELPERS_H
